@@ -126,7 +126,7 @@ static PluginStatus message(PluginLevel level, const char *fmt, ...) {
 
   switch (level) {
   case LDPL_INFO:
-    SyncOut(ctx) << buf;
+    Out(ctx) << buf;
     break;
   case LDPL_WARNING:
     Warn(ctx) << buf;
@@ -176,7 +176,7 @@ static PluginStatus add_input_file(const char *path) {
   Context<E> &ctx = *gctx<E>;
   static i64 file_priority = 100;
 
-  MappedFile<Context<E>> *mf = MappedFile<Context<E>>::must_open(ctx, path);
+  MappedFile *mf = must_open_file(ctx, path);
 
   ObjectFile<E> *file = ObjectFile<E>::create(ctx, mf, "", false);
   ctx.obj_pool.emplace_back(file);
@@ -200,10 +200,7 @@ static PluginStatus release_input_file(const void *handle) {
   LOG << "release_input_file\n";
 
   ObjectFile<E> &file = *(ObjectFile<E> *)handle;
-  if (file.mf->fd != -1) {
-    close(file.mf->fd);
-    file.mf->fd = -1;
-  }
+  file.mf->close_fd();
   return LDPS_OK;
 }
 
@@ -427,7 +424,7 @@ get_api_version(const char *plugin_identifier,
   if (LAPI_V1 < minimal_api_supported)
     Fatal(*gctx<E>) << "LTO plugin does not support V0 or V1 API";
 
-  std::string version = mold_version + "\0"s;
+  std::string version = get_mold_version() + "\0"s;
 
   *linker_identifier = "mold";
   *linker_version = version.data();
@@ -449,7 +446,7 @@ static void load_lto_plugin(Context<E> &ctx) {
     phase = 1;
     gctx<E> = &ctx;
 
-    void *handle = dlopen(ctx.arg.plugin.c_str(), RTLD_NOW | RTLD_GLOBAL);
+    void *handle = dlopen(ctx.arg.plugin.c_str(), RTLD_NOW | RTLD_LOCAL);
     if (!handle)
       Fatal(ctx) << "could not open plugin file: " << dlerror();
 
@@ -570,7 +567,11 @@ static ElfSym<E> to_elf_sym(PluginSymbol &psym) {
 // Returns false if it's GCC.
 template <typename E>
 static bool is_llvm(Context<E> &ctx) {
+#ifdef __MINGW32__
+  return ctx.arg.plugin.ends_with("LLVMgold.dll");
+#else
   return ctx.arg.plugin.ends_with("LLVMgold.so");
+#endif
 }
 
 // Returns true if a given linker plugin supports the get_symbols_v3 API.
@@ -582,25 +583,25 @@ static bool supports_v3_api(Context<E> &ctx) {
 
 template <typename E>
 static PluginInputFile
-create_plugin_input_file(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+create_plugin_input_file(Context<E> &ctx, MappedFile *mf) {
   PluginInputFile file;
-  MappedFile<Context<E>> *mf2 = mf->parent ? mf->parent : mf;
+  MappedFile *mf2 = mf->parent ? mf->parent : mf;
 
   file.name = save_string(ctx, mf2->name).data();
   file.offset = mf->get_offset();
   file.filesize = mf->size;
 
-  if (mf2->fd == -1)
-    mf2->fd = open(file.name, O_RDONLY);
+  mf2->reopen_fd(file.name);
+
   file.fd = mf2->fd;
 
-  if (file.fd == -1)
+  if (!file.fd)
     Fatal(ctx) << "cannot open " << file.name << ": " << errno_string();
   return file;
 }
 
 template <typename E>
-ObjectFile<E> *read_lto_object(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+ObjectFile<E> *read_lto_object(Context<E> &ctx, MappedFile *mf) {
   load_lto_plugin(ctx);
 
   // V0 API's claim_file is not thread-safe.
@@ -644,9 +645,8 @@ ObjectFile<E> *read_lto_object(Context<E> &ctx, MappedFile<Context<E>> *mf) {
   // LLVM needs it and takes the ownership of fd. To prevent "too many
   // open files" issue, we close fd only for GCC. This is ugly, though.
   if (!is_llvm(ctx)) {
-    MappedFile<Context<E>> *mf2 = mf->parent ? mf->parent : mf;
-    close(mf2->fd);
-    mf2->fd = -1;
+    MappedFile *mf2 = mf->parent ? mf->parent : mf;
+    mf2->close_fd();
   }
 
   // Create a symbol strtab
@@ -721,7 +721,7 @@ std::vector<ObjectFile<E> *> do_lto(Context<E> &ctx) {
   // given to the LTO backend. Such sections contains code and data for
   // peripherails (typically GPUs).
   for (ObjectFile<E> *file : ctx.objs) {
-    if (!file->is_lto_obj && file->is_gcc_offload_obj) {
+    if (file->is_alive && !file->is_lto_obj && file->is_gcc_offload_obj) {
       PluginInputFile pfile = create_plugin_input_file(ctx, file->mf);
       int claimed = false;
       claim_file_hook(&pfile, &claimed);
@@ -746,7 +746,7 @@ void lto_cleanup(Context<E> &ctx) {
 
 using E = MOLD_TARGET;
 
-template ObjectFile<E> *read_lto_object(Context<E> &, MappedFile<Context<E>> *);
+template ObjectFile<E> *read_lto_object(Context<E> &, MappedFile *);
 template std::vector<ObjectFile<E> *> do_lto(Context<E> &);
 template void lto_cleanup(Context<E> &);
 
