@@ -1,6 +1,5 @@
 #include "mold.h"
 #include "../common/archive-file.h"
-#include "../common/cmdline.h"
 #include "../common/output-file.h"
 
 #include <cstring>
@@ -22,12 +21,18 @@
 # include <unistd.h>
 #endif
 
+#ifdef MOLD_X86_64
+int main(int argc, char **argv) {
+  return mold::elf::elf_main<mold::elf::X86_64>(argc, argv);
+}
+#endif
+
 namespace mold::elf {
 
 // Read the beginning of a given file and returns its machine type
 // (e.g. EM_X86_64 or EM_386).
 template <typename E>
-std::string_view get_machine_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+std::string_view get_machine_type(Context<E> &ctx, MappedFile *mf) {
   auto get_elf_type = [&](u8 *buf) -> std::string_view {
     bool is_le = (((ElfEhdr<I386> *)buf)->e_ident[EI_DATA] == ELFDATA2LSB);
     bool is_64;
@@ -83,13 +88,15 @@ std::string_view get_machine_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
   case FileType::GCC_LTO_OBJ:
     return get_elf_type(mf->data);
   case FileType::AR:
-    for (MappedFile<Context<E>> *child : read_fat_archive_members(ctx, mf))
-      if (get_file_type(ctx, child) == FileType::ELF_OBJ)
+    for (MappedFile *child : read_fat_archive_members(ctx, mf))
+      if (FileType ty = get_file_type(ctx, child);
+          ty == FileType::ELF_OBJ || ty == FileType::GCC_LTO_OBJ)
         return get_elf_type(child->data);
     return "";
   case FileType::THIN_AR:
-    for (MappedFile<Context<E>> *child : read_thin_archive_members(ctx, mf))
-      if (get_file_type(ctx, child) == FileType::ELF_OBJ)
+    for (MappedFile *child : read_thin_archive_members(ctx, mf))
+      if (FileType ty = get_file_type(ctx, child);
+          ty == FileType::ELF_OBJ || ty == FileType::GCC_LTO_OBJ)
         return get_elf_type(child->data);
     return "";
   case FileType::TEXT:
@@ -101,7 +108,7 @@ std::string_view get_machine_type(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 
 template <typename E>
 static void
-check_file_compatibility(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+check_file_compatibility(Context<E> &ctx, MappedFile *mf) {
   std::string_view target = get_machine_type(ctx, mf);
   if (target != ctx.arg.emulation)
     Fatal(ctx) << mf->name << ": incompatible file type: "
@@ -109,7 +116,7 @@ check_file_compatibility(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 }
 
 template <typename E>
-static ObjectFile<E> *new_object_file(Context<E> &ctx, MappedFile<Context<E>> *mf,
+static ObjectFile<E> *new_object_file(Context<E> &ctx, MappedFile *mf,
                                       std::string archive_name) {
   static Counter count("parsed_objs");
   count++;
@@ -121,12 +128,12 @@ static ObjectFile<E> *new_object_file(Context<E> &ctx, MappedFile<Context<E>> *m
   file->priority = ctx.file_priority++;
   ctx.tg.run([file, &ctx] { file->parse(ctx); });
   if (ctx.arg.trace)
-    SyncOut(ctx) << "trace: " << *file;
+    Out(ctx) << "trace: " << *file;
   return file;
 }
 
 template <typename E>
-static ObjectFile<E> *new_lto_obj(Context<E> &ctx, MappedFile<Context<E>> *mf,
+static ObjectFile<E> *new_lto_obj(Context<E> &ctx, MappedFile *mf,
                                   std::string archive_name) {
   static Counter count("parsed_lto_objs");
   count++;
@@ -140,25 +147,25 @@ static ObjectFile<E> *new_lto_obj(Context<E> &ctx, MappedFile<Context<E>> *mf,
   file->is_in_lib = ctx.in_lib || (!archive_name.empty() && !ctx.whole_archive);
   file->is_alive = !file->is_in_lib;
   if (ctx.arg.trace)
-    SyncOut(ctx) << "trace: " << *file;
+    Out(ctx) << "trace: " << *file;
   return file;
 }
 
 template <typename E>
 static SharedFile<E> *
-new_shared_file(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+new_shared_file(Context<E> &ctx, MappedFile *mf) {
   check_file_compatibility(ctx, mf);
 
   SharedFile<E> *file = SharedFile<E>::create(ctx, mf);
   file->priority = ctx.file_priority++;
   ctx.tg.run([file, &ctx] { file->parse(ctx); });
   if (ctx.arg.trace)
-    SyncOut(ctx) << "trace: " << *file;
+    Out(ctx) << "trace: " << *file;
   return file;
 }
 
 template <typename E>
-void read_file(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+void read_file(Context<E> &ctx, MappedFile *mf) {
   if (ctx.visited.contains(mf->name))
     return;
 
@@ -172,7 +179,7 @@ void read_file(Context<E> &ctx, MappedFile<Context<E>> *mf) {
     return;
   case FileType::AR:
   case FileType::THIN_AR:
-    for (MappedFile<Context<E>> *child : read_archive_members(ctx, mf)) {
+    for (MappedFile *child : read_archive_members(ctx, mf)) {
       switch (get_file_type(ctx, child)) {
       case FileType::ELF_OBJ:
         ctx.objs.push_back(new_object_file(ctx, child, mf->name));
@@ -208,36 +215,47 @@ void read_file(Context<E> &ctx, MappedFile<Context<E>> *mf) {
 
 template <typename E>
 static std::string_view
-deduce_machine_type(Context<E> &ctx, std::span<std::string> args) {
-  for (std::string_view arg : args)
-    if (!arg.starts_with('-'))
-      if (auto *mf = MappedFile<Context<E>>::open(ctx, std::string(arg)))
+detect_machine_type(Context<E> &ctx, std::vector<std::string> paths) {
+  std::erase(paths, "-");
+
+  for (const std::string &path : paths)
+    if (auto *mf = open_file(ctx, path))
+      if (get_file_type(ctx, mf) != FileType::TEXT)
         if (std::string_view target = get_machine_type(ctx, mf);
             !target.empty())
           return target;
+
+  for (const std::string &path : paths)
+    if (auto *mf = open_file(ctx, path))
+      if (get_file_type(ctx, mf) == FileType::TEXT)
+        if (std::string_view target = get_script_output_type(ctx, mf);
+            !target.empty())
+          return target;
+
   Fatal(ctx) << "-m option is missing";
 }
 
 template <typename E>
-MappedFile<Context<E>> *open_library(Context<E> &ctx, std::string path) {
-  MappedFile<Context<E>> *mf = MappedFile<Context<E>>::open(ctx, path);
+MappedFile *open_library(Context<E> &ctx, std::string path) {
+  MappedFile *mf = open_file(ctx, path);
   if (!mf)
     return nullptr;
 
   std::string_view target = get_machine_type(ctx, mf);
-  if (target.empty() || target == E::target_name)
-    return mf;
-  Warn(ctx) << path << ": skipping incompatible file " << target
-            << " " << (int)E::e_machine;
-  return nullptr;
+  if (!target.empty() && target != E::target_name) {
+    Warn(ctx) << path << ": skipping incompatible file: " << target
+              << " (e_machine " << (int)E::e_machine << ")";
+    return nullptr;
+  }
+  return mf;
 }
 
 template <typename E>
-MappedFile<Context<E>> *find_library(Context<E> &ctx, std::string name) {
+MappedFile *find_library(Context<E> &ctx, std::string name) {
   if (name.starts_with(':')) {
     for (std::string_view dir : ctx.arg.library_paths) {
       std::string path = std::string(dir) + "/" + name.substr(1);
-      if (MappedFile<Context<E>> *mf = open_library(ctx, path))
+      if (MappedFile *mf = open_library(ctx, path))
         return mf;
     }
     Fatal(ctx) << "library not found: " << name;
@@ -246,22 +264,22 @@ MappedFile<Context<E>> *find_library(Context<E> &ctx, std::string name) {
   for (std::string_view dir : ctx.arg.library_paths) {
     std::string stem = std::string(dir) + "/lib" + name;
     if (!ctx.is_static)
-      if (MappedFile<Context<E>> *mf = open_library(ctx, stem + ".so"))
+      if (MappedFile *mf = open_library(ctx, stem + ".so"))
         return mf;
-    if (MappedFile<Context<E>> *mf = open_library(ctx, stem + ".a"))
+    if (MappedFile *mf = open_library(ctx, stem + ".a"))
       return mf;
   }
   Fatal(ctx) << "library not found: " << name;
 }
 
 template <typename E>
-MappedFile<Context<E>> *find_from_search_paths(Context<E> &ctx, std::string name) {
-  if (MappedFile<Context<E>> *mf = MappedFile<Context<E>>::open(ctx, name))
+MappedFile *find_from_search_paths(Context<E> &ctx, std::string name) {
+  if (MappedFile *mf = open_file(ctx, name))
     return mf;
 
   for (std::string_view dir : ctx.arg.library_paths)
-    if (MappedFile<Context<E>> *mf =
-        MappedFile<Context<E>>::open(ctx, std::string(dir) + "/" + name))
+    if (MappedFile *mf =
+        open_file(ctx, std::string(dir) + "/" + name))
       return mf;
   return nullptr;
 }
@@ -294,7 +312,7 @@ static void read_input_files(Context<E> &ctx, std::span<std::string> args) {
     } else if (arg == "--end-lib") {
       ctx.in_lib = false;
     } else if (remove_prefix(arg, "--version-script=")) {
-      MappedFile<Context<E>> *mf = find_from_search_paths(ctx, std::string(arg));
+      MappedFile *mf = find_from_search_paths(ctx, std::string(arg));
       if (!mf)
         Fatal(ctx) << "--version-script: file not found: " << arg;
       parse_version_script(ctx, mf);
@@ -308,11 +326,11 @@ static void read_input_files(Context<E> &ctx, std::span<std::string> args) {
         state.back();
       state.pop_back();
     } else if (remove_prefix(arg, "-l")) {
-      MappedFile<Context<E>> *mf = find_library(ctx, std::string(arg));
+      MappedFile *mf = find_library(ctx, std::string(arg));
       mf->given_fullpath = false;
       read_file(ctx, mf);
     } else {
-      read_file(ctx, MappedFile<Context<E>>::must_open(ctx, std::string(arg)));
+      read_file(ctx, must_open_file(ctx, std::string(arg)));
     }
   }
 
@@ -340,7 +358,7 @@ int elf_main(int argc, char **argv) {
 
   // If no -m option is given, deduce it from input files.
   if (ctx.arg.emulation.empty())
-    ctx.arg.emulation = deduce_machine_type(ctx, file_args);
+    ctx.arg.emulation = detect_machine_type(ctx, file_args);
 
   // Redo if -m is not x86-64.
   if constexpr (is_x86_64<E>)
@@ -364,7 +382,7 @@ int elf_main(int argc, char **argv) {
     on_complete = fork_child();
 #endif
 
-  acquire_global_lock(ctx);
+  acquire_global_lock();
 
   tbb::global_control tbb_cont(tbb::global_control::max_allowed_parallelism,
                                ctx.arg.thread_count);
@@ -416,6 +434,9 @@ int elf_main(int argc, char **argv) {
   // "Kill" .eh_frame input sections after symbol resolution.
   kill_eh_frame_sections(ctx);
 
+  // Split mergeable section contents into section pieces.
+  split_section_pieces(ctx);
+
   // Resolve mergeable section pieces to merge them.
   resolve_section_pieces(ctx);
 
@@ -470,6 +491,10 @@ int elf_main(int argc, char **argv) {
   // Bin input sections into output sections.
   create_output_sections(ctx);
 
+  // Handle --section-align options.
+  if (!ctx.arg.section_align.empty())
+    apply_section_align(ctx);
+
   // Add synthetic symbols such as __ehdr_start or __end.
   add_synthetic_symbols(ctx);
 
@@ -520,8 +545,7 @@ int elf_main(int argc, char **argv) {
 
   // If .ctors/.dtors are to be placed to .init_array/.fini_array,
   // we need to reverse their contents.
-  if (ctx.has_init_array && ctx.has_ctors)
-    fixup_ctors_in_init_array(ctx);
+  fixup_ctors_in_init_array(ctx);
 
   // Handle --shuffle-sections
   if (ctx.arg.shuffle_sections != SHUFFLE_SECTIONS_NONE)
@@ -682,10 +706,11 @@ int elf_main(int argc, char **argv) {
 
   std::cout << std::flush;
   std::cerr << std::flush;
+
   if (on_complete)
     on_complete();
 
-  release_global_lock(ctx);
+  release_global_lock();
 
   if (ctx.arg.quick_exit)
     _exit(0);
@@ -696,12 +721,6 @@ int elf_main(int argc, char **argv) {
   return 0;
 }
 
-#ifdef MOLD_X86_64
-int main(int argc, char **argv) {
-  return elf_main<X86_64>(argc, argv);
-}
-#endif
-
 using E = MOLD_TARGET;
 
 template int elf_main<E>(int, char **);
